/* -*- mode: js2; js2-basic-offset: 4; indent-tabs-mode: nil -*- */
/*
 * Magnifier Preferences - GnomeShell Magnifier settings via GSettings
 *
 * Copyright (C) 2010, 2011 Inclusive Design Research Centre
 * Contributor: Joseph Scheuhammer <clown@alum.mit.edu>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

const Clutter = imports.gi.Clutter;
const Gdk = imports.gi.Gdk;
const Gio = imports.gi.Gio;
const GLib = imports.gi.GLib;
const Gtk = imports.gi.Gtk;

const DBus = imports.dbus;
const Gettext = imports.gettext;
const Lang = imports.lang;
const Signals = imports.signals;

const SHOW_KEY                      = 'screen-magnifier-enabled';
const SCREEN_POSITION_KEY           = 'screen-position';
const MAG_FACTOR_KEY                = 'mag-factor';
const LENS_MODE_KEY                 = 'lens-mode';
const CLAMP_MODE_KEY                = 'scroll-at-edges';
const MOUSE_MODE_KEY                = 'mouse-tracking';

const INVERT_BRIGHTNESS_KEY         = 'invert-brightness';
const BRIGHTNESS_KEY                = 'brightness';
const CONTRAST_KEY                  = 'contrast';

const CROSSHAIRS_SHOW_KEY           = 'show-cross-hairs';
const CROSSHAIRS_CLIP_KEY           = 'cross-hairs-clip';
const CROSSHAIRS_COLOR_KEY          = 'cross-hairs-color';
const CROSSHAIRS_THICKNESS_KEY      = 'cross-hairs-thickness';
const CROSSHAIRS_OPACITY_KEY        = 'cross-hairs-opacity';
const CROSSHAIRS_LENGTH_KEY         = 'cross-hairs-length';

/**
 * Magnifier preferences dialog.
 */

function MagnifierPreferences(uiFile) {
    this._init(uiFile);
}

MagnifierPreferences.prototype = {
    _init: function(uiFile) {

        // Calculate the viewport for the 'docked' screen position.
        // note:  arbitrarily situated at left edge of screen, half way down;
        // 2/3 the width of the screen, and 1/3 the height.
        let screenWidth = Gdk.Screen.width();
        let screenHeight = Gdk.Screen.height();
        let top = screenHeight / 3 * 2;
        let portHeight = screenHeight / 3;
        let portWidth = screenWidth / 2;
        this._dockedViewPort = [0, top, portWidth, top + portHeight];

        // Initialize DBus connection to the magnifier service.
        DBus.proxifyPrototype(MagnifierDBus.prototype, magnifierIface);
        this._magDBusAccess = new MagDBusAccess('/org/gnome/Magnifier');

        // Initialize GSettings
        this._appSettings = new Gio.Settings({ schema: 'org.gnome.desktop.a11y.applications' });
        this._magSettings = new Gio.Settings({ schema: 'org.gnome.desktop.a11y.magnifier' });

        // Initialize state of the widgets in the UI.
        let builder = new Gtk.Builder();
        builder.add_from_file(uiFile);

        // Set the widgets according to the current stored GSettings values.
        // Show/hide magnifier toggle ...
        let magOnRadio = builder.get_object('magOnRadio');
        let magOffRadio = builder.get_object('magOffRadio');
        let aSetting = this._appSettings.get_boolean(SHOW_KEY);
        magOnRadio.set_active(aSetting);
        magOffRadio.set_active(!aSetting);
        this._appSettings.connect('changed::' + SHOW_KEY, function(settings) {
            let toggle = settings.get_boolean(SHOW_KEY);
            magOnRadio.set_active(toggle);
            magOffRadio.set_active(!toggle);
        });

        // Mouse tracking mode radio buttons ...
        let radioGroup = builder.get_object('none').get_group();
        this._updateRadioGroup(this._magSettings, MOUSE_MODE_KEY, radioGroup);
        this._magSettings.connect('changed::' + MOUSE_MODE_KEY, Lang.bind(this, this._updateRadioGroup, radioGroup));

        // Screen position radio ...
        this._dockedRadio = builder.get_object('docked');
        radioGroup = this._dockedRadio.get_group();
        this._updateRadioGroup(this._magSettings, SCREEN_POSITION_KEY, radioGroup);
        this._magSettings.connect('changed::' + SCREEN_POSITION_KEY, Lang.bind(this, this._updateRadioGroup, radioGroup));

        // Magnification factor slider ...
        let magSlider = builder.get_object('magFactorSlider');
        aSetting = this._magSettings.get_double(MAG_FACTOR_KEY);
        magSlider.set_value(aSetting);
        this._magSettings.connect('changed::' + MAG_FACTOR_KEY, function(settings) {
            magSlider.set_value(settings.get_double(MAG_FACTOR_KEY));
        });

        // Lens mode checkbox ...
        let lensMode = builder.get_object('moveableLensCheckbox');
        aSetting = this._magSettings.get_boolean(LENS_MODE_KEY);
        lensMode.set_active(aSetting);
        this._magSettings.connect('changed::' + LENS_MODE_KEY, function(settings) {
            lensMode.set_active(settings.get_boolean(LENS_MODE_KEY));
        });

        // ... Clamp scrolling at screen edges checkbox ...
        let edgeScrollMode = builder.get_object('scrollAtEdges');
        aSetting = this._magSettings.get_boolean(CLAMP_MODE_KEY);
        edgeScrollMode.set_active(aSetting);
        this._magSettings.connect('changed::' + CLAMP_MODE_KEY, function(settings) {
            edgeScrollMode.set_active(settings.get_boolean(CLAMP_MODE_KEY));
        });

        // ... Cross hairs: show/hide ...
        let showCrosshairsMode = builder.get_object('xHairsShowHide');
        aSetting = this._magSettings.get_boolean(CROSSHAIRS_SHOW_KEY);
        showCrosshairsMode.set_active(aSetting);
        this._magSettings.connect('changed::' + CROSSHAIRS_SHOW_KEY, function(settings) {
            showCrosshairsMode.set_active(settings.get_boolean(CROSSHAIRS_SHOW_KEY));
        });

        // ... Cross hairs: color ...
        let crosshairsColor = builder.get_object('xHairsPicker');
        aSetting = this._magSettings.get_string(CROSSHAIRS_COLOR_KEY);
        let [parsed, gdkColor] = Gdk.color_parse(aSetting);
        crosshairsColor.set_color(gdkColor);
        this._magSettings.connect('changed::' + CROSSHAIRS_COLOR_KEY, function(settings) {
            let [ok, color] = Gdk.color_parse(settings.get_string(CROSSHAIRS_COLOR_KEY));
            crosshairsColor.set_color(color);
        });

        // ... Cross hairs: opacity ...
        let crosshairsOpacity = builder.get_object('xHairsOpacitySlider');
        aSetting = this._magSettings.get_double(CROSSHAIRS_OPACITY_KEY);
        crosshairsOpacity.set_value(aSetting);
        this._magSettings.connect('changed::' + CROSSHAIRS_OPACITY_KEY, function(settings) {
            crosshairsOpacity.set_value(settings.get_double(CROSSHAIRS_OPACITY_KEY));
        });

        // ... Cross hairs: thickness ...
        let crosshairsThickness = builder.get_object('xHairsThicknessSpinner');
        aSetting = this._magSettings.get_int(CROSSHAIRS_THICKNESS_KEY);
        crosshairsThickness.set_value(aSetting);
        this._magSettings.connect('changed::' + CROSSHAIRS_THICKNESS_KEY, function(settings) {
            crosshairsThickness.set_value(settings.get_int(CROSSHAIRS_THICKNESS_KEY));
        });

        // ... Cross hairs: clip ...
        let clipCrosshairsMode = builder.get_object('xHairsClipCheckbox');
        aSetting = this._magSettings.get_boolean(CROSSHAIRS_CLIP_KEY);
        clipCrosshairsMode.set_active(aSetting);
        this._magSettings.connect('changed::' + CROSSHAIRS_CLIP_KEY, function(settings) {
            clipCrosshairsMode.set_active(settings.get_boolean(CROSSHAIRS_CLIP_KEY));
        });

        // ... Cross hairs: length
        let crosshairsLength = builder.get_object('xHairsLengthSlider');
        aSetting = this._magSettings.get_int(CROSSHAIRS_LENGTH_KEY);
        crosshairsLength.set_value(aSetting);
        this._magSettings.connect('changed::' + CROSSHAIRS_LENGTH_KEY, function(settings) {
            crosshairsLength.set_value(settings.get_int(CROSSHAIRS_LENGTH_KEY));
        });

        // ... Brightness ...
        let invertBrightness = builder.get_object('invertBrightnessCheckbox');
        aSetting = this._magSettings.get_boolean(INVERT_BRIGHTNESS_KEY);
        invertBrightness.set_active(aSetting);
        this._magSettings.connect('changed::' + INVERT_BRIGHTNESS_KEY, function(settings) {
            invertBrightness.set_active(settings.get_boolean(INVERT_BRIGHTNESS_KEY));
        });
        this._brightRed = builder.get_object('brightRedSpinner');
        this._brightGreen = builder.get_object('brightGreenSpinner');
        this._brightBlue = builder.get_object('brightBlueSpinner');
        aSetting = this._magSettings.get_string(BRIGHTNESS_KEY);
        this._updateColorSpinners(aSetting, this._brightRed, this._brightGreen, this._brightBlue);
        this._magSettings.connect('changed::' + BRIGHTNESS_KEY, Lang.bind(this, function() {
            let newBrightness = this._magSettings.get_string(BRIGHTNESS_KEY);
            this._updateColorSpinners(newBrightness, this._brightRed, this._brightGreen, this._brightBlue);
        }));

        // ... Contrast.
        this._contrastRed = builder.get_object('contrastRedSpinner');
        this._contrastGreen = builder.get_object('contrastGreenSpinner');
        this._contrastBlue = builder.get_object('contrastBlueSpinner');
        aSetting = this._magSettings.get_string(CONTRAST_KEY);
        this._updateColorSpinners(aSetting, this._contrastRed, this._contrastGreen, this._contrastBlue);
        this._magSettings.connect('changed::' + CONTRAST_KEY, Lang.bind(this, function() {
            let newContrast = this._magSettings.get_string(CONTRAST_KEY);
            this._updateColorSpinners(newContrast, this._contrastRed, this._contrastGreen, this._contrastBlue);
        }));

        // Attach our widget handlers.
        builder.connect_signals_full(Lang.bind(this, this._gtkConnectSignals));

        // Handle closing.
        let mainWindow = builder.get_object('magPrefsDialog');
        let closeButton = builder.get_object('closeButton');
        closeButton.connect('clicked', function() {
            mainWindow.destroy();
        });
        mainWindow.connect('delete_event', function() {
            return false;
        });
        mainWindow.connect('destroy', function() {
            Gtk.main_quit();
        });
        mainWindow.show_all();
    },

    doShowHide: function(radioButton) {
        this._appSettings.set_boolean(SHOW_KEY, radioButton.get_active());
    },

    doMouseTracking: function(radioButton) {
        if (radioButton.get_active())
            this._magSettings.set_string(MOUSE_MODE_KEY, radioButton.get_name());
    },

    doScreenPosition: function(radioButton) {
        if (radioButton.get_active()) {
            if (radioButton == this._dockedRadio)
                this._magDBusAccess.moveFirstZoomRegion(this._dockedViewPort);
            else
                this._magSettings.set_string(SCREEN_POSITION_KEY, radioButton.get_name());
        }
    },

    doMagnification: function(slider) {
        // two decimal places
        let magFactor = parseFloat(slider.get_value().toFixed(2));
        this._magSettings.set_double(MAG_FACTOR_KEY, magFactor);
    },

    doLensMode: function(checkbox) {
        let lensMode = checkbox.get_active();
        let isDocked = this._dockedRadio.get_active();
        this._magSettings.set_boolean(LENS_MODE_KEY, lensMode);
        if (!lensMode && isDocked) {
            this._magDBusAccess.moveFirstZoomRegion(this._dockedViewPort);
        }
    },

    doClampScrolling: function(checkbox) {
        this._magSettings.set_boolean(CLAMP_MODE_KEY, checkbox.get_active());
    },

    doShowCrosshairs: function(checkbox) {
        this._magSettings.set_boolean(CROSSHAIRS_SHOW_KEY, checkbox.get_active());
    },

    doCrosshairsColor: function(colorButton) {
        // colorVal is in the form '#rrrrggggbbbb'.  Convert to '#rrggbb'.
        let colorVal = colorButton.get_color().to_string();
        let cVal = '#';
        cVal += colorVal.substring(1,3);      // red
        cVal += colorVal.substring(5,7);      // green
        cVal += colorVal.substring(9,11);     // blue
        this._magSettings.set_string(CROSSHAIRS_COLOR_KEY, cVal);
    },

    doCrosshairsOpacity: function(opacitySlider) {
        let opacity = opacitySlider.get_value();
        this._magSettings.set_double(CROSSHAIRS_OPACITY_KEY, opacity);
    },

    doCrosshairsThickness: function(thicknessSpinner) {
        this._magSettings.set_int(CROSSHAIRS_THICKNESS_KEY, thicknessSpinner.get_value());
    },

    doCrosshairsClip: function(checkbox) {
        this._magSettings.set_boolean(CROSSHAIRS_CLIP_KEY, checkbox.get_active());
    },

    doCrosshairsLength: function(lengthSlider) {
        this._magSettings.set_int(CROSSHAIRS_LENGTH_KEY, lengthSlider.get_value());
    },

    doInvertBrightness: function(checkbox) {
        this._magSettings.set_boolean(INVERT_BRIGHTNESS_KEY, checkbox.get_active());
    },

    doBrightness: function(spinner) {
        let brightness = new Clutter.Color();
        brightness.red = this._brightRed.get_value() * 255;
        brightness.green = this._brightGreen.get_value() * 255;
        brightness.blue = this._brightBlue.get_value() * 255;
        let bString = brightness.to_string().substring(0,7); // drop alpha-bits.
        this._magSettings.set_string(BRIGHTNESS_KEY, bString);
    },

    doContrast: function(spinner) {
        let contrast = new Clutter.Color();
        contrast.red = this._contrastRed.get_value() * 255;
        contrast.green = this._contrastGreen.get_value() * 255;
        contrast.blue = this._contrastBlue.get_value() * 255;
        let cString = contrast.to_string().substring(0,7); // drop alpha-bits.
        this._magSettings.set_string(CONTRAST_KEY, cString);
    },

    // Function for Gkt.Builder to use for connecting the dialog's widget events
    // to their handlers.
    _gtkConnectSignals: function(builder, widget, signal, handlerName) {
        if (this[handlerName]) {
            widget.connect(signal, Lang.bind(this, this[handlerName]));
        }
    },

    // Update the states of the radio buttons in a group for the given settings
    // key.
    _updateRadioGroup: function(settings, key, radioGroup) {
        let value = settings.get_string(key);
        for (let i = 0; i < radioGroup.length; i++) {
            let aRadio = radioGroup[i];
            let radioName = aRadio.get_name();
            if (radioName == value)
                aRadio.set_active(true);
            else
                aRadio.set_active(false);
        }
    },

    // Update color adjustments.
    _updateColorSpinners: function(colorString, redSpinner, greenSpinner, blueSpinner) {
        let color = new Clutter.Color();
        color.from_string(colorString);
        redSpinner.set_value(color.red / 255);
        greenSpinner.set_value(color.green / 255);
        blueSpinner.set_value(color.blue / 255);
    }
}
Signals.addSignalMethods(MagnifierPreferences.prototype);

// DBus interfaces to the magnifier and its zoom regions
let magnifierIface = {
    name: 'org.gnome.Magnifier',
    methods: [
        { name: 'getZoomRegions', inSignature: '', outSignature: 'ao' }
    ]
};

let zoomRegionIface = {
    name: 'org.gnome.Magnifier.ZoomRegion',
    methods: [
        { name: 'moveResize', inSignature: 'ai', outSignature: '' }
    ]
};

// Magnifier remote DBus proxy
function MagnifierDBus(objectPath) {
    this._init(objectPath);
}

MagnifierDBus.prototype = {
    _init: function(objectPath) {
        DBus.session.proxifyObject(this, 'org.gnome.Magnifier', objectPath);
    }
};

// Object to access the magnifier and zoom region DBbus services.
function MagDBusAccess() {
    this._init();
}

MagDBusAccess.prototype = {
    _init:  function() {
        this._magDBus = new MagnifierDBus('/org/gnome/Magnifier');
        this._firstZoomRegionPath = null;
    },

    /**
     * moveFirstZoomRegion:
     * Acquire the first zoom region, and move it to the given location.
     * @viewPort:   The rectangle defining the location and size of the zoom
     *              region, in the form [left, top, right bottom].
     */
    moveFirstZoomRegion: function(viewPort) {
        // The actual first zoom region may have been destroyed and replaced
        // since the last invocation.  Get the current first zoom region.
        this._firstZoomRegionPath = null;
        this._magDBus.getZoomRegionsRemote(
            Lang.bind(this, this.getZoomRegionsCallback, viewPort)
        );
    },

    /**
     * getZoomRegionsCallback:
     * Callback that remote Magnifier dbus object calls to pass back the object
     * paths of its zoom regions.  If any are passed, then record the path of
     * the first zoom region, and then call the method that actually moves it.
     * @result:     Array of ZoomRegion object paths.
     * @excp:       DBus exception.
     * @viewPort:   Array [left, top, right, bottom].
     */
    getZoomRegionsCallback: function(result, excp, viewPort) {
        if (result && result.length > 0) {
            this._firstZoomRegionPath = result[0];
            this.doMoveZoomRegion(viewPort);
        }
    },

    /**
     * doMoveZoomRegion:
     * Move the first zoom region, if any, to the given location.
     * @viewPort:  An array of the form [left, top, right, bottom] defining
     *             where to move the zoom region to.
     */
    doMoveZoomRegion: function(viewPort) {
        if (this._firstZoomRegionPath == null)
            return;
        let zoomRegionProxy = new MagnifierDBus(this._firstZoomRegionPath);
        if (zoomRegionProxy) {
            DBus.proxifyPrototype(zoomRegionProxy, zoomRegionIface);
            zoomRegionProxy.moveResizeRemote(viewPort);
        }
    }
};

function main(params) {
    if (params['progName'])
        GLib.set_prgname(params['progName']);
    if (params['localeDir'])
        Gettext.bindtextdomain('gnome-shell', params['localeDir']);

    Gtk.init(null, null);
    let magPrefs = new MagnifierPreferences(params['uiFile']);
    magPrefs.connect('destroy', function() { Gtk.main_quit(); });
    Gtk.main();
}
